;; Generated package description from gnugo.el  -*- no-byte-compile: t -*-
(define-package "gnugo" "3.1.1.0.20201129.144920" "play GNU Go in a buffer" '((ascii-art-to-unicode "1.5") (xpm "1.0.1") (cl-lib "0.5")) :authors '(("Thien-Thi Nguyen" . "ttn@gnu.org")) :maintainer '("Thien-Thi Nguyen" . "ttn@gnu.org") :keywords '("games" "processes") :url "http://www.gnuvola.org/software/gnugo/")
